#include <stdio.h>
#include <stdlib.h>   /* EXIT_FAILURE,exit */
#include <string.h>
#include <ctype.h>
#include "kh.h"

#define BUFSIZE_INP (1024 * 16)
#define BUFSIZE_OUT (1024 * 32)

typedef unsigned char uchar; /* Хʸ */

typedef struct 
{
  uchar  *inp_str;     /* entry */
  int     inp_len;     /* length of entry */
  uchar  *out_str;     /* output of entry */
  int     out_len;     /* length of output */
  int     cnt;
  int     done;        /* for generating a vocabulary list */
  int     last_pos;
  uchar   last_byte;
} TYPE_DIC;

static int readdic(uchar *p_file_name);
static int line_count(uchar *p_file_name);
static int make_index();
static int translate(uchar *p_inp_name);
static int dicgen(uchar *p_out_str, uchar *p_inp_str, int p_inp_len);
static int lookup(uchar *p_out_str, uchar *p_inp_str, int p_inp_len);
static TYPE_DIC *getdic(uchar *p_str);
int main(int argc, char **argv);
static int destroy_dic(void);
int iskanji(unsigned char c);

uchar *input_name   = NULL;   /* input filename */
uchar *dic_name  = NULL;     /* dictionary filename */

int dic_cnt = 0;             /* the number of entry in a dictionary */
TYPE_DIC *dic_tbl = NULL;   /* pointer of dictionary table */
int dic_idx[256 + 1];        /* index table of a dictionary */
int dic_len[256];            /* length of entry in a dictionary */

/* Option */
int opt_all_possible = FALSE; /* all possible patterns */
int opt_back         = FALSE; /* back and forward accessment */
int opt_dic          = FALSE; /* output dictionary data */
int opt_xml_tag      = FALSE; /* add a caridge return after '>' tag */

#if defined(SJIS)
int opt_sjis         = TRUE; /* SJIS option */
int opt_euc          = FALSE;  /* EUC option (default) */
#else
int opt_sjis         = FALSE; /* SJIS option */
int opt_euc          = TRUE;  /* EUC option (default) */
#endif

///////////////////////////////////////////////////////////////////////
//
//  read the contents of dictionary from a file.
//
//////////////////////////////////////////////////////////////////////
static int readdic(uchar *p_file_name)
{
  int i, j, k;
  int w_len;
  int w_lens;
  
  FILE *w_file;
  int w_dic_cnt;
  uchar w_str[1024];
  uchar *w_dup;
  TYPE_DIC *w_dic;
  
  /* count the number of entry. */
  if (dic_cnt == 0) {
    w_dic_cnt = line_count(p_file_name);
    if (w_dic_cnt < 0) {
      fprintf(stderr, "kh: %s: No such file or directory\n", p_file_name);
      return(-1);
    }
  } else {
    w_dic_cnt = dic_cnt;
  }
  
  /* memory allocations */
  dic_tbl = (TYPE_DIC *)malloc(w_dic_cnt * sizeof(TYPE_DIC));
  if (dic_tbl == NULL) {
    fprintf(stderr, "kh: %s: No such file or directory\n", p_file_name);
    return(-1);
  }
  memset(dic_tbl, 0x00, w_dic_cnt * sizeof(TYPE_DIC));
  
  /* open the dictionary file. */
  w_file = fopen(p_file_name, "r");
  if (w_file == NULL) {
    fprintf(stderr, "kh: %s: No such file or directory\n", p_file_name);
    return -1;
  }
  
  /* parse each entry in a dictionary */
  for ( i = 0; i < w_dic_cnt; ) {
    if (fgets(w_str, sizeof(w_str), w_file) == NULL) break;
    if (w_str[0] == '#') continue;  /* a comment line */ 
    if (w_str[0] == '\n') continue; /* an only carridge return line */ 
    
    /* editing an entry */
    w_lens = strlen(w_str);
    for (j = 0; j < w_lens; j++ ) {
      /* ignore data after spaces, '#', and '\n' */
      if (w_str[j] == ' ' || w_str[j] == '#' || w_str[j] == '\n' ) {
        w_lens = j; /* formal length of the strings found */
        break;
      }
    } 
    w_len  = w_lens ; 
    w_str[w_len] = 0x00; /* replace '\n' with '\0'. */

    for (j = 1; j < w_len; j++) {
      if (w_str[j] == '=') {             /* in the case of a delimiter  */
        w_dup  = malloc(w_lens);         /* allocate memory of an entry */
        if (w_dup == NULL) {
          perror("malloc");
          return(-1);
        }
        memcpy(w_dup, w_str, w_lens);    /* copy w_str to w_dup */
        w_dup[j] = 0x00;                 /* replace '=' with '\0' */
        
        w_dic = &dic_tbl[i];      /* get an address i of dictionary table  */
        w_dic->inp_str   = w_dup;  /* copy an address of entry to the table */
        w_dic->inp_len   = j;
        w_dic->last_byte = w_dup[ w_dic->last_pos = j - 1 ];
        
        j++;
        w_dic->out_str = &w_dup[j];    /* put the address of the rest */
        w_dic->out_len = w_len - j;
        w_dic->cnt     = FALSE;
        j = -1;
        break;
      }
    }
    
    if (j > 0) {
      fprintf(stderr, "warning: %s\n", w_str);
      continue;
    }
    i++;
  }
  fclose(w_file);

  dic_cnt = i; /* set the number of entry in a dictionary. */
  make_index(); /* then make index for search. */

  return 0;

}


///////////////////////////////////////////////////////////////////////
//
//  count the number of entry in a dictionary
//
//////////////////////////////////////////////////////////////////////
static int line_count(uchar *p_file_name)
{
  FILE *w_file;
  int  w_cnt;
  char w_str[1024];
  
  w_file = fopen(p_file_name, "r");
  if (w_file == NULL)
    return(-1);

  w_cnt = 0;  
  while (fgets(w_str, sizeof(w_str), w_file)) {
    if (w_str[0] != '#')  /* do not count lines for comments.  */
      w_cnt++;
  }
  
  if (w_cnt < 1) {        /* no valid lines */
    fprintf(stderr,"The dictionary, %s, is empty.\n", dic_name);
    exit(0);
  }
  fclose(w_file);
  return(w_cnt);
}


///////////////////////////////////////////////////////////////////////
//
//   make index
//
//////////////////////////////////////////////////////////////////////
static int make_index()
{
  int i, j;
  int w_byte, w_len;
  int w_to;
  
  memset(dic_idx, 0x00, sizeof(dic_idx));
  memset(dic_len, 0x00, sizeof(dic_len));

  w_len  = dic_tbl[0].inp_len;
  w_byte = dic_tbl[0].inp_str[0];

  for (i = 1; i < dic_cnt; i++) {
    w_byte = dic_tbl[i].inp_str[0];            /* ƬʸΥХ */
    if (dic_tbl[i-1].inp_str[0] != w_byte) {   /* ֹʤǤ   */
      dic_idx[w_byte] = i;                     /* ΥХȤϿ */
      dic_len[ dic_tbl[i-1].inp_str[0] ] = w_len;
      w_len = dic_tbl[i].inp_len;
      continue;
    }
    
    if (w_len > dic_tbl[i].inp_len)
      w_len = dic_tbl[i].inp_len;
  }
  
  dic_idx[256] = dic_cnt;
  dic_len[w_byte] = w_len;
  
  w_to = dic_tbl[0].inp_str[0];
  for (i = 255; i > w_to; i--) {
    if (dic_idx[i] == 0)
      dic_idx[i] = dic_idx[i + 1];
  }


  /* put the serial number of the same entries. */ 
  j = 1;
  dic_tbl[0].cnt = 1;
  for (i = 1; i < dic_cnt; i++) {
    if (dic_tbl[i-1].inp_len == dic_tbl[i].inp_len && 
        memcmp(dic_tbl[i-1].inp_str, dic_tbl[i].inp_str,
               dic_tbl[i].inp_len) == 0) {
      j++;
      dic_tbl[i].cnt = j;
    } else {
      j = 1;
      dic_tbl[i].cnt = 1;
    }
  }
  return 0;
}

///////////////////////////////////////////////////////////////////////
//
//  translate 
//
//////////////////////////////////////////////////////////////////////
static int translate(uchar *p_inp_name)
{
  int     i, j, k;
  FILE   *w_inp_file;
  FILE   *w_out_file;
  uchar  *w_inp_str;
  uchar  *w_out_str;
  int     w_inp_len;
  int     w_out_len;
  
  static uchar w_inp_vbuf[BUFSIZ];
//  static uchar w_out_vbuf[BUFSIZ];
  
  uchar   w_str[BUFSIZE_OUT];
  int     w_prev, w_next;
  int     w_form_len;
  
  if (p_inp_name == NULL) {
    w_inp_file = stdin;
  } else {
    w_inp_file = fopen(p_inp_name, "r");
    if (w_inp_file == NULL) {
      fprintf(stderr, "Couldn't open input file %s\n", p_inp_name);
      return(-1);
    }
  }
  
  w_out_file = stdout;
  
  /* memory allocations */
  w_inp_str = malloc(BUFSIZE_INP);
  w_out_str = malloc(BUFSIZE_OUT);
  if (w_inp_str == NULL || w_out_str == NULL) {
    perror("malloc");
    return(-1);
  }

  setvbuf(w_inp_file, w_inp_vbuf, _IOLBF, BUFSIZ);
  
  while (fgets(w_inp_str, BUFSIZ, w_inp_file)) {
    if (w_inp_str[0] == '#' ) /* comment line. skip it */
      continue; 

    if (!opt_dic) {
      w_out_len = lookup(w_out_str, w_inp_str, strlen(w_inp_str));
    
      for (i = 0; i < w_out_len; i++) /* ڤʸԤѴ */
        if (w_out_str[i] == ';' || w_out_str[i] == '\t') 
          w_out_str[i] = '\n';
    }

    if (opt_dic) {
      w_out_len = dicgen(w_out_str, w_inp_str, strlen(w_inp_str));
      for (i = 0; i < w_out_len; i++) /* ʸԤѴ */
        if (w_out_str[i] == '\t') 
          w_out_str[i] = '\n';
    }
    fwrite(w_out_str, w_out_len, 1, w_out_file);
  }
  fflush(w_out_file);
  if (w_inp_file != stdin) 
    fclose(w_inp_file);

  free(w_inp_str);
  free(w_out_str);

  return 0;
}

//////////////////////////////////////////////////////////////////////
//
//  dictionary regenerator
//
//////////////////////////////////////////////////////////////////////
static int dicgen(uchar *p_out_str, uchar *p_inp_str, int p_inp_len)
{
  int     i, j, k;
  uchar   tmp_str[BUFSIZ];
  TYPE_DIC *w_dic;
  TYPE_DIC *w_dic_tmp;
    
  for (i = 0, j = 0; i < p_inp_len; ) {

    /* look it up with the dicationary */
    w_dic = getdic(&p_inp_str[i]);
    
    /* if something is in the dictionary, convert it */
    if (w_dic != NULL) { 
      w_dic_tmp = w_dic;

      if (w_dic->cnt > 1 && opt_all_possible) { /* output all the patterns */
        for (k = 0; k < w_dic->cnt; k++) {
          if (w_dic_tmp->done != TRUE) {
            memcpy(&p_out_str[j], w_dic_tmp->inp_str, w_dic_tmp->inp_len);
            j += w_dic_tmp->inp_len;
            memcpy(&p_out_str[j], "=", 1);
            j++;
            memcpy(&p_out_str[j], w_dic_tmp->out_str, w_dic_tmp->out_len);
            j += w_dic_tmp->out_len;
            memcpy(&p_out_str[j], "\n", 1);
            j++;
            w_dic_tmp->done = TRUE;
            w_dic_tmp--;
          }
        }
      } else {
        if (w_dic_tmp->done != TRUE) {
          memcpy(&p_out_str[j], w_dic_tmp->inp_str, w_dic_tmp->inp_len);
          j += w_dic_tmp->inp_len;
          memcpy(&p_out_str[j], "=", 1);
          j++;
          memcpy(&p_out_str[j], w_dic_tmp->out_str, w_dic_tmp->out_len);
          j += w_dic_tmp->out_len;
          memcpy(&p_out_str[j], "\n", 1);
          j++;
          w_dic_tmp->done = TRUE;
        }
      }
      
      i += w_dic->inp_len;
      continue;
    }
    
    /* cannot find it in the dictionary */
    if (p_inp_str[i]&0x80) {   /* 2 bytes */
      i += 2;
    } else {                   /* 1 byte for number tags */
      i++;
    }
  }
  return j;
}

//////////////////////////////////////////////////////////////////////
//
//  refer, replace, and rearrange
//
// KKS0753  ȤϤơᤤơޤ
// KKS0759  ˤˡᴢˤˡˤ
//
//////////////////////////////////////////////////////////////////////
static int lookup(uchar *p_out_str, uchar *p_inp_str, int p_inp_len)
{
  int     i, j, k;
  uchar   tmp_str[BUFSIZ];
  TYPE_DIC *w_dic;
  TYPE_DIC *w_dic_tmp;
    
  for (i = 0, j = 0; i < p_inp_len; ) {

    /* look it up with the dicationary */
    w_dic = getdic(&p_inp_str[i]);
    
    /* if something is in the dictionary, convert it */
    if (w_dic != NULL) { 
      w_dic_tmp = w_dic;

      if (w_dic->cnt > 1 && opt_all_possible) { /* output all the patterns */
        for (k = 0; k < w_dic->cnt; k++) {
          memcpy(&p_out_str[j], "---\n", 4);
          j += 4;
          memcpy(&p_out_str[j], w_dic_tmp->out_str, w_dic_tmp->out_len);
          j += w_dic_tmp->out_len;
          w_dic_tmp--;
        }
        memcpy(&p_out_str[j], "---\n", 4);
        j += 4;
      } else {
        memcpy(&p_out_str[j], w_dic_tmp->out_str, w_dic_tmp->out_len);
        j += w_dic_tmp->out_len;
      }
      i += w_dic->inp_len;
      continue;
    }
    
    /* cannot find it in the dictionary */
//    if (p_inp_str[i]&0x80) {   /* 2 bytes */
    if (iskanji(p_inp_str[i])) {   /* 2 bytes */
      memcpy(&p_out_str[j], &p_inp_str[i], 2);
      j += 2;
      memcpy(&p_out_str[j], " missing\n", 9);
      j += 9;
      i += 2;
    } else if (p_inp_str[i] == '>' && opt_xml_tag) { /* add a caridge return */
        memcpy(&p_out_str[j], ">\n", 2);
        j += 2;
        i++;
    } else {                   /* 1 byte for number tags */
      memcpy(&p_out_str[j], &p_inp_str[i], 1);
      j++;
      i++;
    }
  }
  return j;
}

//////////////////////////////////////////////////////////////////////
//
//  pick up a p_str from a dictionary
//
//////////////////////////////////////////////////////////////////////
static TYPE_DIC *getdic(uchar *p_str)
{
  int i;
  int w_pos;
  int w_len, w_cmp;
  int w_from, w_to;
  int w_start, w_stop;
  int w_cnt;

  // θϡw_from  w_to Υǥåɥ쥹ˤ롣
  w_from = dic_idx[ p_str[0] ];
  w_to   = dic_idx[ p_str[0] + 1] - 1;

  if (w_from > w_to) 
    return NULL;
  
  w_len   = dic_len[ p_str[0] ];
  w_start = -1;
  w_stop  = w_to;

  for (;;) {
    if (w_from > w_to) break;
    w_pos = (w_from+w_to) >> 1;
    w_cmp = memcmp(p_str, dic_tbl[w_pos].inp_str, w_len);

    if (w_cmp < 0) {
      w_to   = w_pos - 1;
      w_stop = w_to;
    } else if (w_cmp > 0) {
      w_from = w_pos + 1;
    } else /* if (w_cmp == 0) */ {
      w_start = w_pos;
      w_to    = w_pos - 1;
    }
  }

  if (w_start == -1)
    return NULL;
  
  if (dic_tbl[w_start].inp_len > w_len)
    w_len++;
  
  w_from = w_start;
  w_to   = w_stop;
  w_stop = -1;
  for (;;) {
    if (w_from > w_to) break;
    w_pos = (w_from + w_to) >> 1;
    w_cmp = memcmp(p_str, dic_tbl[w_pos].inp_str, w_len);

    if (w_cmp < 0) {
      w_to = w_pos - 1;
    } else if (w_cmp > 0) {
      w_from  = w_pos + 1;
      w_start = w_from;
    } else /* if (w_cmp == 0) */ {
      w_stop = w_pos;
      w_from = w_pos + 1;
    }
  }
  
  if (w_stop == -1) {
    return NULL;
  }
  
  /* pick up a longest matched pattern from the dictionary */
  w_cnt = 0;
  for (i = w_stop; i >= w_start; i -= dic_tbl[i].cnt) {
    if (p_str[dic_tbl[i].last_pos] == dic_tbl[i].last_byte &&
        memcmp(p_str, dic_tbl[i].inp_str, dic_tbl[i].inp_len) == 0 ) {
      w_pos = i;
      w_cnt = dic_tbl[i].cnt;
//      printf("test w_cnt %s %d\n", dic_tbl[i].inp_str, w_cnt);
      break;
    }
  }
  
  switch(w_cnt) {
  case 0:                       
    return NULL;
  default:
    return(&dic_tbl[w_pos]); /* return the address of the entry. */
  }
}

///////////////////////////////////////////////////////////////////////
//
//  main: get options, read dictionary, an interface to a translate function
//
//////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
  int i, j, r;
  char *op;
    
  dic_name = getenv("KHDIC");
  if (dic_name == NULL) 
    dic_name = "./kh.dic";

  for (argv++; *argv && argv[0][0] == '-' && argv[0][1]; argv++) {
    for (op = &argv[0][1]; *op; op++) {
      switch (*op) {
      case 'b':
        opt_back = TRUE;
        break;
      case 'e':
        opt_sjis = FALSE;
        opt_euc  = TRUE;
        break;
      case 's':
        opt_sjis = TRUE;
        opt_euc  = FALSE;
        break;
      case 'x':
        opt_xml_tag = TRUE;
        break;
      case 'm':
        opt_all_possible = TRUE;
        break;
      case 'd':
        opt_dic= TRUE;
        break;
      case 'f':
        if (!*++argv) {
          fprintf(stderr,"The option -%c requires an argument.\n", *op);
          exit(0);
        }
        dic_name = *argv;
        break;
      case 'v':
        fprintf(stderr,"%s\n",VERSION);
        exit(0);
      case 'h':
        fprintf(stderr,"%s\n",USAGE);
        exit(0);
      default:
        fprintf(stderr,"%% %s: invalid option -%c\n", PROG_NAME, *op);
        exit(0);
      }
    }
  }

  r = readdic(dic_name);
  if (r < 0)
    exit(-1);

  if (*argv == NULL)
    r = translate(NULL);
  else
    for (;*argv;argv++)
      r = translate(*argv);
  if (r < 0) 
    exit(-1);

  destroy_dic();
  return 0;
}

static int destroy_dic(void)
{
  free(dic_tbl);
  return 0;
}

int iskanji(unsigned char c)
{
  if (opt_sjis)
    return((0x81 <= c && c <= 0x9F) || (0xE0 <= c && c <= 0xFC)); /* SJIS */
  if (opt_euc)
    return(0x8E <= c && c <= 0xFE && 0x8F != c && c != 0xA0);     /* EUC  */
}
